# This Python file uses the following encoding: utf-8

# Copyright (C) 2011 Pablo Sebastián Carranza Giotto.
#  
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# #####################################################
#
# This program is a port and update of rtve-mediateca-dl to python
# It covers the same functionality but adds some eye candy and adds
# some more features.
# Thanks to Sergio García-Cuevas González for creating the original
# bash version of mediateca downloader
#
# #####################################################
# 
# website: https://bitbucket.org/jimcerberus/rtvedownloader/
# 
# #####################################################

import sys
import urllib
import re
import os.path
from optparse import OptionParser
from xml.etree.ElementTree import ElementTree
import traceback


def main():
    parser = OptionParser(prog='rtvedownload',
                      version='0.0.1',
                      usage='%prog [options] url',
                      description='Download a video or podcast from TVE "A la carta": ' + 
                        'http://www.rtve.es/alacarta/tve/');
                      
    parser.add_option("-f", 
                      "--file", 
                      dest="filename",
                      help="save video/podcast to FILE", 
                      metavar="FILE")
    
    parser.add_option("-v",
                      "--verbose",
                      dest="verbose",
                      action="store_true",
                      help="verbose mode") 
    
    parser.add_option("-n",
                      "--no-action",
                      dest="noaction",
                      action="store_true",
                      help="does not download any file, just checks for availability") 
                      
    (options, args) = parser.parse_args()
    if not args:
        print_message('\nNo url defined\n\n')
        parser.print_help()
        sys.exit(1)
    
    if options.filename and len(args) > 1:
        print_message('Could not set filename when downloading more than 1 file\n\n')
        parser.print_help()
        sys.exit(1)
    
    downloader = RTVEDownloader(options)
    for url in args:
        try:
            value = downloader.download(url)
            (downloaded, message) = value
            print_message(message)
            if not downloaded:
                print_message('\nMaybe should retry %s?\n\n' % url)
                
        except Exception, e:
            print_message('Unexpected exception:\n\n %s' % e)
            traceback.print_exc(file=sys.stdout)
            
            
    sys.exit(0)
    

class RTVEDownloader:
    
    DATA_URI_TEMPLATE = 'http://www.rtve.es/swf/data/%s/%s/%s/%s/%s.xml'
    ASSET_URI_TEMPLATE = '%s/ASSET_DATA_%s/%s/ASSET_DATA_%s-%s.xml'
    RESOURCE_URI_TEMPLATE = "http://www.rtve.es/resources/TE_NGVA/%s"
    
    KBYTE = 1024
    MBYTE = KBYTE * 1024
    
    def __init__(self, options):
        self.options = options
        self.url_opener = self.create_url_opener()
        
        
    def output(self, message, force=False):
        if self.options.verbose or force:
            print_message(message)

    
    def download(self, url):
        self.output('Downloading file from url %s' % url)
        
        params = self.get_id_url(url)
        if not params:
            return (False, 'Could not get flash params')

        self.output('Obtained id %s' % params['id'])
        
        asset_params = self.build_asset_template_params(params)
        
        if not asset_params or 'url' not in asset_params.keys():
            return (False, 'Could not build template url')
        
        self.output('Obtained template url %s' % asset_params['url'])
        
        asset = self.get_asset_url(asset_params)
        if not asset:
            return (False, 'Could not build asset url')

        self.output('Obtained asset url %s' % asset)
        
        resource_fields = self.get_resource_url(asset)
        if not resource_fields:
            return (False, 'Could not get resource fields')
        
        filename = self.download_file(resource_fields)
        
        return (True, 'File %s downloaded' % filename)
    
    
    def get_id_url(self, url):
        content = self.urlopen(url)
        if not content:
            raise Exception('Url %s does not exists' % url)
        
        try:
            for line in content:
                found = re.search(u'name=\"flashvars\" value=\"[A-z0-9&;=_]*\"', line)
                if found:
                    s = str(found.group())
                    found = re.search(u'value=\".*\"', s)
                    s = str(found.group())
                    s = s[7:len(s) - 1]
                    s = s.replace('&amp;', '&')
                    key_values = s.split('&')
                    params = {}
                    for key_value in key_values:
                        (key, value) = key_value.split('=')
                        params[key] = value
                        
                    params['id'] = params['assetID'].split('_')[0]
                    return params
                    
            return None
        
        finally:
            if content:
                content.close()
             

    def build_asset_template_params(self, params):
        (id, lang, path) = params['assetID'].split('_')
        type = path[:len(path) - 1]
        return {'id': id, 
                'lang' : lang, 
                'path' : path, 
                'type' : type, 
                'url' : RTVEDownloader.DATA_URI_TEMPLATE % (lang, path, type, self.scramble_id(id), id)}
    
    
    def scramble_id(self, id):
        i = len(id)
        return '%s/%s/%s/%s' % (id[i-1], id[i-2], id[i-3], id[i-4])
    
    
    def urlopen(self, url):
        return self.url_opener.open(url)
    
    
    def load_params_dict(self, params):
        dparams= {}
        for line in params.split(','):
            key, value = line.split('::')
            dparams[key] = value
        
        return dparams
    
    
    def get_asset_url(self, asset_params):
        template_url = self.urlopen(asset_params['url'])
        try:
            if not template_url:
                raise Exception('Resource %s does not exists, not compatible? please report a bug' % template_url)
            
            etree = ElementTree()
            etree.parse(template_url)
            nodes = etree.getroot().findall('plugins/plugin')
            for node in reversed(nodes):
                name = node.attrib['name']
                if not name == 'multicdn':
                    continue
                params = node.attrib['params']
                dparams = self.load_params_dict(params)
                assetDataId = dparams['assetDataId']
                type = asset_params['type'].upper()
                path = dparams['defaultPath']
                return RTVEDownloader.ASSET_URI_TEMPLATE % (path, type, self.scramble_id(assetDataId), type, assetDataId)
            
            return None
        
        finally:
            template_url.close()
    
    
    def get_resource_url(self, url):
        content = self.urlopen(url)
        try:
            if not content:
                raise Exception('Resource %s does not exists, not compatible? please report a bug' % url)

            etree = ElementTree()
            etree.parse(content)
            
            root = etree.getroot()
            location = root.attrib['defaultLocation']
            
            filename = os.path.basename(location)
            extension = filename.split('.')[1]
            
            ibase = str(location).find(extension)
            base = location[ibase:]
            
            fields_list = root.findall('*/field')
            try:
                fields = {}
                for field in fields_list:
                    if len(field.getchildren()) != 2: 
                        continue
                    
                    key, value = tuple(field.getchildren())
                    fields[key.text] = unicode(value.text)
                    
                fields['url'] = RTVEDownloader.RESOURCE_URI_TEMPLATE % base
                return fields
            except Exception, e:
                raise Exception('Could not load fields: %s' % e)
        
        finally:
            content.close()


    def download_file(self, resource_fields):
        url = resource_fields['url']
        
        filename = resource_fields['AS_TITLE']
        filename = filename.replace('/', '-').replace('\\', '-')
        
        encoding = sys.stdout.encoding
    
        if encoding == None:
            encoding = 'utf-8'

        if encoding != 'utf-8':
            filename = filename.encode(encoding, 'replace')
            
        extension = os.path.basename(url).split('.')[1]
        
        if self.options.filename:
            filename = self.options.filename
        else:
            cwd = os.getcwd().decode(encoding)
            filename = filename.decode(encoding)
            filename = os.path.join(cwd, filename)
            
        if not filename.endswith(extension):
            filename += '.' + extension
                
        myurl = self.create_url_opener()
        exists_size = 0
        local_params = 'wb'
        if not self.options.noaction:
            if os.path.exists(filename):
                local_params = 'ab'
                exists_size = os.path.getsize(filename)
                myurl.addheader('Range', 'bytes=%s-' % exists_size)
            
        remote_file = myurl.open(url)
        
        size = remote_file.headers['Content-Length']
        size = int(size)
        
        if self.options.noaction:
            self.output('No real action, file %s is available for download, skipping\n' % url, True)
            return filename
        
        if size == 0:
            self.output('File %s is already downloaded' % filename)
            return filename
        
        if exists_size == 0:
            self.output('Downloading file %s %s' % 
                        (filename, 
                         bytes_to_readable(size)), 
                         force=True)
            self.output('\tfrom final url %s' % url)
            
        else:
            self.output('Resuming file %s - starting at %s of %s' % 
                        (filename, 
                         bytes_to_readable(exists_size), 
                         bytes_to_readable(size)), 
                         force=True)
            self.output('\tfrom final url %s' % url)
        
        progress = ProgressBar(min_value=0, max_value=size, mode='fixed')
        progress.update_amount(exists_size)
        
        oldvalue = str(progress)
        
        sys.stdout.write('%s\r' % oldvalue)
        sys.stdout.flush()

        with open(filename, local_params) as local_file:        
            try:
                for line in remote_file:
                    local_file.write(line)
                    progress.increment_amount(len(line))
                    newvalue = str(progress)
                    if newvalue != oldvalue:
                        sys.stdout.write('%s\r' % newvalue) 
                        sys.stdout.flush()
                        oldvalue = newvalue
                    
            finally:
                remote_file.close()
                sys.stdout.write('\n')
        
        return filename
    
    
    def create_url_opener(self):
        return myURLOpener()
    
    
class myURLOpener(urllib.FancyURLopener):
    """Create sub-class in order to overide error 206.  This error means a
       partial file is being sent,
       which is ok in this case.  Do nothing with this error.
    """
    def http_error_206(self, url, fp, errcode, errmsg, headers, data=None):
        pass
    
    def http_error_404(self, url, fp, errcode, errmsg, headers, data=None):
        raise Exception('Url %s does not exists', url)
    
    
    
class ProgressBar:
    
    def __init__(self, min_value = 0, max_value = 100, width=77, **kwargs):
        self.char = kwargs.get('char', '=')
        self.empty_char = kwargs.get('empty_char', '-')
        self.mode = kwargs.get('mode', 'dynamic') # fixed or dynamic
        if not self.mode in ['fixed', 'dynamic']:
            self.mode = 'fixed'
 
        self.bar = ''
        self.min = min_value
        self.max = max_value
        self.span = max_value - min_value
        self.width = width
        self.amount = 0       # When amount == max, we are 100% done 
        self.readable_max = bytes_to_readable(self.max)
        self.update_amount(0)
 
 
    def increment_amount(self, add_amount = 1):
        """
        Increment self.amount by 'add_ammount' or default to incrementing
        by 1, and then rebuild the bar string. 
        """
        new_amount = self.amount + add_amount
        if new_amount < self.min: new_amount = self.min
        if new_amount > self.max: new_amount = self.max
        self.amount = new_amount
        self.build_bar()
 
 
    def update_amount(self, new_amount = None):
        """
        Update self.amount with 'new_amount', and then rebuild the bar 
        string.
        """
        if not new_amount: new_amount = self.amount
        if new_amount < self.min: new_amount = self.min
        if new_amount > self.max: new_amount = self.max
        self.amount = new_amount
        self.build_bar()
 
 
    def build_bar(self):
        """
        Figure new percent complete, and rebuild the bar string base on 
        self.amount.
        """
        diff = float(self.amount - self.min)
        percent_done = int(round((diff / float(self.span)) * 100.0))
 
        # figure the proper number of 'character' make up the bar 
        all_full = self.width - 2
        num_hashes = int(round((percent_done * all_full) / 100))
 
        if self.mode == 'dynamic':
            # build a progress bar with self.char (to create a dynamic bar
            # where the percent string moves along with the bar progress.
            self.bar = self.char * num_hashes
        else:
            # build a progress bar with self.char and spaces (to create a 
            # fixe bar (the percent string doesn't move)
            self.bar = self.char * num_hashes + self.empty_char * (all_full-num_hashes)

        percent_str = str(percent_done) + "%"
        self.bar = '[%s] %s - %s of %s    ' % (self.bar, 
                 percent_str, 
                 bytes_to_readable(self.amount),
                 self.readable_max)
 
 
    def __str__(self):
        return str(self.bar)


def bytes_to_readable(size):
    size = int(size)
    if size < RTVEDownloader.KBYTE:
        return '%dB' % size
    
    # Kbytes
    if size < RTVEDownloader.MBYTE:
        readable_size_k = int(size / RTVEDownloader.KBYTE)
        
        return '%dKB' % readable_size_k
    
    readable_size_m = int(round(float(size) / RTVEDownloader.MBYTE, 0))
    
    return '%dMB' % readable_size_m


def print_message(message):
    encoding = sys.stdout.encoding
    
    if encoding == None:
        encoding = 'utf-8'
        
    if not message.endswith('\n'):
        message += '\n'
        
    sys.stdout.write(message.encode(encoding))

    
if __name__ == '__main__':
    main()