#!/usr/bin/env python
# -*- coding: utf8 -*-

from io import open, StringIO
from mock import Mock, MagicMock
from rtvedownloader import RTVEDownloader
from test import BaseTest
import os
import unittest
import random

class URLTest(BaseTest):

    def setUp(self):
        BaseTest.setUp(self)
        self.params = {'assetID': '843192_es_videos', 
                  'playerId': 'player843192', 
                  'location': 'alacarta', 
                  'id': '843192'}
        
        self.asset_params = { 'lang': 'es', 
                             'path': 'videos', 
                             'url': 'http://www.rtve.es/swf/data/es/videos/video/2/9/1/3/843192.xml', 
                             'type': 'video', 
                             'id': '843192' }
        
        self.asset_url = 'http://www.rtve.es/scd/CONTENTS/ASSET_DATA_VIDEO/4/0/8/3/ASSET_DATA_VIDEO-523804.xml'
        
        self.resource_fields = {
                    'url' : 'http://www.rtve.es/resources/TE_NGVA/flv/2/6/1280770006662.flv',
                    'AS_TITLE' : 'Documentos TV - La caza del último nazi' }
        self.filename = 'file.flv'
    
    
    def test_get_id_url_returns_params(self):
        real = RTVEDownloader(self.options)
        f = os.path.join(os.path.dirname(__file__), 'res/get_id.html')
        real.urlopen = Mock(return_value=open(f, encoding='UTF8'))
        params = real.get_id_url(self.url)
        
        real.urlopen.assert_called_once_with(self.url)
        self.assertEqual(params, self.params)


    def test_get_id_url_with_void_file_returns_none(self):
        real = RTVEDownloader(self.options)
        f = os.path.join(os.path.dirname(__file__), 'res/void.html')
        real.urlopen = Mock(return_value=open(f, encoding='UTF8'))
        params = real.get_id_url(self.url)
        
        real.urlopen.assert_called_once_with(self.url)
        self.assertEqual(params, None)
        
        
    def test_build_asset_template_params(self):
        real = RTVEDownloader(self.options)
        asset_params = real.build_asset_template_params(self.params)
        self.assertEquals(asset_params, self.asset_params)
        
        
    def test_get_asset_url(self):
        real = RTVEDownloader(self.options)
        f = os.path.join(os.path.dirname(__file__), 'res/asset_params.xml')
        real.urlopen = Mock(return_value=open(f, encoding='UTF8'))
        asset_url = real.get_asset_url(self.asset_params)
        
        real.urlopen.assert_called_once_with(self.asset_params['url'])
        self.assertEqual(self.asset_url, asset_url)
        
    
    def test_get_asset_url_with_void_xml_returns_none(self):
        real = RTVEDownloader(self.options)
        f = os.path.join(os.path.dirname(__file__), 'res/void_asset_params.xml')
        real.urlopen = Mock(return_value=open(f, encoding='UTF8'))
        asset_url = real.get_asset_url(self.asset_params)
        
        real.urlopen.assert_called_once_with(self.asset_params['url'])
        self.assertEqual(asset_url, None)


    def test_get_asset_url_raises_exception_when_template_is_none(self):
        real = RTVEDownloader(self.options)
        real.urlopen = Mock(return_value=None)
        
        self.assertRaises(Exception, real.get_asset_url, self.asset_params)
        real.urlopen.assert_called_once_with(self.asset_params['url'])
        
    
    def test_get_id_url_gets_called_when_download(self):
        real = RTVEDownloader(self.options)
        mock = Mock(return_value=None)
        real.get_id_url = mock
        
        (result, message) = real.download(self.url)
        
        mock.assert_called_once_with(self.url)
        
        self.assertFalse(result)
        self.assertFalse(message is None)


    def test_get_resource_url_raises_exception_when_no_content(self):
        real = RTVEDownloader(self.options)
        real.urlopen = Mock(return_value=None)
        
        self.assertRaises(Exception, real.get_resource_url, self.url)
        real.urlopen.assert_called_once_with(self.url)
        

    def test_get_resource_url_with_valid_file(self):
        real = RTVEDownloader(self.options)
        f = os.path.join(os.path.dirname(__file__), 'res/resource.xml')
        real.urlopen = Mock(return_value=open(f, encoding='UTF8'))
        
        fields = real.get_resource_url(self.url)
        
        real.urlopen.assert_called_once_with(self.url)
        
        self.assertTrue('url' in fields)
        self.assertTrue('AS_TITLE' in fields)
        
    
    def test_get_id_url_with_no_content_throws_exception(self):
        real = RTVEDownloader(self.options)
        mock = Mock(return_value=None)
        real.urlopen = mock
        
        self.assertRaises(Exception, real.download, self.url)
        
        mock.assert_called_once_with(self.url)
        
        
    def test_get_resource_fields(self):
        real = RTVEDownloader(self.options)
        
        get_id_mock = Mock(return_value=self.params)
        real.get_id_url = get_id_mock
        
        build_asset_template_params_mock = Mock(return_value=None)
        real.build_asset_template_params = build_asset_template_params_mock
        
        (result, message) = real.download(self.url)
        
        self.assertFalse(result)
        self.assertFalse(message is None)
        get_id_mock.assert_called_once_with(self.url)
        build_asset_template_params_mock.assert_called_once_with(self.params)
        
        
    def test_get_asset_url_returns_none(self):
        real = RTVEDownloader(self.options)
        step1 = Mock(return_value=self.params)
        step2 = Mock(return_value=self.asset_params)
        step3 = Mock(return_value=None)

        real.get_id_url = step1
        real.build_asset_template_params = step2
        real.get_asset_url = step3
        
        (result, message) = real.download(self.url)
        
        self.assertFalse(result)
        self.assertFalse(message is None)

        step1.assert_called_once_with(self.url)
        step2.assert_called_once_with(self.params)
        step3.assert_called_once_with(self.asset_params)
        
    
    def test_get_resource_fields_returns_none(self):
        # step1 : params = self.get_id_url(url)
        # step2 : asset_params = self.build_asset_template_params(params)
        # step3 : asset = self.get_asset_url(asset_params)
        # step4 : resource_fields = self.get_resource_url(asset)
        
        options = self.options
        options.verbose = True
        
        real = RTVEDownloader(options)
        step1 = Mock(return_value=self.params)
        step2 = Mock(return_value=self.asset_params)
        step3 = Mock(return_value=self.asset_url)
        step4 = Mock(return_value=None)
        
        real.get_id_url = step1
        real.build_asset_template_params = step2
        real.get_asset_url = step3
        real.get_resource_url = step4
        
        (result, message) = real.download(self.url)
        
        self.assertFalse(result)
        self.assertFalse(message is None)
        
        step1.assert_called_once_with(self.url)
        step2.assert_called_once_with(self.params)
        step3.assert_called_once_with(self.asset_params)
        step4.assert_called_once_with(self.asset_url)

        
        
    def test_full_process(self):
        # step1 : params = self.get_id_url(url)
        # step2 : asset_params = self.build_asset_template_params(params)
        # step3 : asset = self.get_asset_url(asset_params)
        # step4 : resource_fields = self.get_resource_url(asset)
        # step5 : filename = self.download_file(resource_fields)
        
        options = self.options
        options.verbose = True
        
        real = RTVEDownloader(options)
        step1 = Mock(return_value=self.params)
        step2 = Mock(return_value=self.asset_params)
        step3 = Mock(return_value=self.asset_url)
        step4 = Mock(return_value=self.resource_fields)
        step5 = Mock(return_value=self.filename)
        
        real.get_id_url = step1
        real.build_asset_template_params = step2
        real.get_asset_url = step3
        real.get_resource_url = step4
        real.download_file = step5
        
        (result, message) = real.download(self.url)
        
        self.assertTrue(result)
        self.assertFalse(message is None)
        self.assertTrue(self.filename in message)
        
        step1.assert_called_once_with(self.url)
        step2.assert_called_once_with(self.params)
        step3.assert_called_once_with(self.asset_params)
        step4.assert_called_once_with(self.asset_url)
        step5.assert_called_once_with(self.resource_fields)


class FileDownloadTest(BaseTest):
    
    def setUp(self):
        BaseTest.setUp(self)
        self.options.verbose = True
        self.options.noaction = False
        self.options.filename = os.path.join(os.path.dirname(__file__), 'res/fake_file.flv')

        self.resource_fields = {
                    'url' : 'http://fake_url.flv',
                    'AS_TITLE' : 'fake_file' }


    def test_download_file_no_action(self):
        real = RTVEDownloader(self.options)
        real.options.noaction = True
        
        mockFile = Mock()
        mockFile.headers = {'Content-Length' : '1'}
        
        urlopener = Mock()
        urlopener.open = Mock(return_value=mockFile)
        
        createopener = Mock(return_value=urlopener)
        real.create_url_opener = createopener
        
        filename = real.download_file(self.resource_fields)
        
        self.assertEquals(filename, self.options.filename)
        createopener.assert_called_once_with()
        urlopener.open.assert_called_once_with(self.resource_fields['url'])

    
    def test_download_file_with_size_0(self):
        real = RTVEDownloader(self.options)
        
        mockFile = Mock()
        mockFile.headers = {'Content-Length' : '0'}
        
        urlopener = Mock()
        urlopener.open = Mock(return_value=mockFile)
        
        createopener = Mock(return_value=urlopener)
        real.create_url_opener = createopener
                
        filename = real.download_file(self.resource_fields)
        
        self.assertFalse(os.path.isfile(filename))
        
        self.assertEquals(filename, self.options.filename)
        createopener.assert_called_once_with()
        urlopener.open.assert_called_once_with(self.resource_fields['url'])
    
    
    def test_download_file_with_2048_bytes_in_two_lines(self):
        real = RTVEDownloader(self.options)
        real.options.filename = None
        
        def create_content():
            s = []
            for i in range(1024):
                s.append(str(random.choice('abcdefghijklmnopqrstuvwxyz0123456789')))
            return ''.join(s)

        proposed_content = [create_content(), create_content()]
        
        mock_file = MagicMock()
        mock_file.__iter__.return_value = iter(proposed_content)
        mock_file.headers = {'Content-Length' : 2048}
        
        urlopener = Mock()
        urlopener.open = Mock(return_value=mock_file)
        
        createopener = Mock(return_value=urlopener)
        real.create_url_opener = createopener
        
        filename = real.download_file(self.resource_fields)
        
        self.options.filename = filename
        
        self.assertTrue(os.path.isfile(filename))
        self.assertEquals(os.path.getsize(filename), 2048)
        
        createopener.assert_called_once_with()
        urlopener.open.assert_called_once_with(self.resource_fields['url'])
        
        with open(filename, 'br') as f:
            for content in proposed_content:
                file_content = f.read(1024)
                self.assertTrue(file_content, content)

    

    def test_download_file_with_1_byte(self):
        real = RTVEDownloader(self.options)
        
        proposed_content = 'a'
        
        mock_file = MagicMock()
        mock_file.__iter__.return_value = iter([proposed_content])
        mock_file.headers = {'Content-Length' : '1'}
        
        urlopener = Mock()
        urlopener.open = Mock(return_value=mock_file)
        
        createopener = Mock(return_value=urlopener)
        real.create_url_opener = createopener
        
        filename = real.download_file(self.resource_fields)
        
        self.assertTrue(os.path.isfile(filename))
        self.assertEquals(os.path.getsize(filename), 1)
        
        self.assertEquals(filename, self.options.filename)
        createopener.assert_called_once_with()
        urlopener.open.assert_called_once_with(self.resource_fields['url'])
        
        with open(filename, 'br') as f:
            file_content = f.read()
        
        self.assertEquals(file_content, proposed_content)
                
    def test_download_file_with_2_bytes_using_resume(self):
        real = RTVEDownloader(self.options)
        
        proposed_content = 'a'
        
        mock_file = MagicMock()
        mock_file.__iter__.return_value = iter([proposed_content])
        mock_file.headers = {'Content-Length' : '1'}
        
        urlopener = Mock()
        urlopener.open = Mock(return_value=mock_file)
        urlopener.addheader = Mock()
        
        createopener = Mock(return_value=urlopener)
        real.create_url_opener = createopener
        
        filename = real.download_file(self.resource_fields)
        
        self.assertTrue(os.path.isfile(filename))
        self.assertEquals(os.path.getsize(filename), 1)
        
        self.assertEquals(filename, self.options.filename)
        createopener.assert_called_once_with()
        urlopener.open.assert_called_once_with(self.resource_fields['url'])
        
        mock_file.__iter__.return_value = iter([proposed_content])
        
        filename = real.download_file(self.resource_fields)

        self.assertTrue(os.path.isfile(filename))
        self.assertEquals(os.path.getsize(filename), 2)

        self.assertEquals(filename, self.options.filename)
        createopener.assert_called_with()
        urlopener.open.assert_called_with(self.resource_fields['url'])
        urlopener.addheader.assert_called_once_with('Range', 'bytes=1-')
        
        self.assert_sysout_has('Resuming')
        
        with open(filename, 'br') as f:
            file_content = f.read()
        
        self.assertEquals(file_content, proposed_content + proposed_content)

        
    
    def tearDown(self):
        BaseTest.tearDown(self)
        if self.options.filename:
            if os.path.isfile(self.options.filename):
                os.remove(self.options.filename)
            

if __name__ == "__main__":
    unittest.main()
